#!/usr/bin/ruby -w

if RUBY_VERSION > '1.8.7'
  $major = ""
else
  $major = "#{RUBY_VERSION.split('.')[0..1].join('.')}"
end

require File.join(File.dirname(__FILE__), 'rubyreq')

if RUBY_VERSION <= '1.8.7'
require 'rubynode'
end

class DynStr < RuntimeError; end

class String
  attr_accessor :node
end

def string_parse(node)
  case node.type
  when :str
    node.nd_lit
  when :call
    case node.nd_mid
    when :<<, :+
      string_parse(node.nd_recv) + string_parse(node.nd_args.nd_1st)
    else
      raise DynStr
    end
  else
    raise DynStr
  end
end

def string(node)
  str = string_parse(node)
  str.node = node
  str
rescue DynStr => e
  node
end

def get_arg(node, arg)
if RUBY_VERSION <= '1.8.7'
  return nil unless node.class == RubyNode
else
  return nil unless node.class == RubyVM::Node
end
  return nil unless node.type == :array
  return nil if node.nd_alen < arg

  (arg - 1).times{ node = node.nd_next }

  node.nd_head
end

def parse(node)
  req = []
  perh_req = []
  cur_req = req
  
  stack = [node]

  while not stack.empty?
    node = stack.pop

    cur_req = req if node == :req
if RUBY_VERSION <= '1.8.7'
    next unless node.kind_of? RubyNode
else
    next unless node.kind_of? RubyVM::Node
end

    case node.type
    when :newline
      stack.push node.nd_next
    when :fcall
      cur_req << case node.nd_mid
      when :require, :load
        string get_arg(node.nd_args, 1)
      when :autoload
        string get_arg(node.nd_args, 2)
      end
    when :call
      cur_req << case node.nd_recv.nd_vid
      when :Kernel
        case node.nd_mid
        when :require, :load
          string get_arg(node.nd_args, 1)
        end
      else
        string get_arg(node.nd_args, 2) if node.nd_mid == :autoload
      end
    when :module, :class, :sclass
if RUBY_VERSION <= '1.8.7'
      stack.push node.nd_body.nd_next
else
      stack.push node.nd_body
end
    when :block
      stack.push node.nd_next, node.nd_head
    else
if RUBY_VERSION > '1.8.7' && node.type == :scope
      stack.push node.nd_body
else
      if cur_req.equal? req
        stack.push :req
        cur_req = perh_req
      end
      stack.push node.u3_node, node.u2_node, node.u1_node
end
    end
  end

  return req, perh_req
end

def process_file(f)
  prefix, name = RubyReq::explode f
  ext = File.extname(f)

  if ext != '.so'
if RUBY_VERSION <= '1.8.7'
    raise "#{f}: syntax check failed" unless RubyReq::system($ruby, '-c', f) == 0
else
    raise SyntaxError, "#{f}: syntax check failed" unless RubyReq::system($ruby, '-c', f) == 0
end
  end
  if prefix and name
    prefix = RubyReq::split_canon(prefix.sub(/^#{RubyReq::root}/, ''))[0]
    $weakprov_path.each do |w|
      if name.index(w) == 0 and name[w.length].chr == File::SEPARATOR
        file = name[w.length+1..-1]
        basename, ext = RubyReq::base_ext(file)
        $weak_prov[file] = f unless ext == '.rb'
        $weak_prov[basename] = f if ['.rb', '.so'].include? ext
        break
      end
    end
  else
    file = RubyReq::split_canon(f.sub(/^#{RubyReq::root}/, ''))[0]
    begin
      file = file[file.index(File::SEPARATOR)+1..-1]
      basename, ext = RubyReq::base_ext(file)
      $weak_prov[file] = f unless ext == '.rb'
      $weak_prov[basename] = f if ['.rb', '.so'].include? ext
    end while file.index(File::SEPARATOR) != nil
  end
  return [] if ext == '.so'

  node = File.read(f).parse_to_nodes
  
  req, perh_req = parse(node)

  perh_req.to_a.flatten.compact.uniq.each do |r|
if RUBY_VERSION > '1.8.7'
    if r.class == RubyVM::Node
      RubyReq.warn "#{f}:#{r.line} perhaps depends on module pointed by dynamic string"
    else
      RubyReq.warn "#{f}:#{r.node.line} perhaps depends on 'ruby(#{r})'"
    end
else
    if r.class == RubyNode
      RubyReq.warn "#{f}:#{r.line} perhaps depends on module pointed by dynamic string"
    else
      RubyReq.warn "#{f}:#{r.node.line} perhaps depends on 'ruby#{$major}(#{r})'"
    end
end
  end

  req.to_a.flatten.compact.delete_if do |r|
if RUBY_VERSION <= '1.8.7'
    if r.class == RubyNode
      RubyReq.warn "#{f}:#{r.line} depends on module pointed by dynamic string"
      true
    end
else
    if r.class == RubyVM::Node
      RubyReq.warn "#{f}:#{r.line} depends on module pointed by dynamic string"
      true
    end
end
  end
end

req = Array.new
$weak_prov = Hash.new
$ruby = ENV['RPM_RUBY'] || '/usr/bin/ruby'
$weakprov_path = (ENV['RPM_RUBY_WEAKPROV_PATH'] || '').split.sort { |a,b| b.length <=> a.length }

if RUBY_VERSION <= '1.8.7'
RubyReq::argv.each do |f|
  begin
    req += process_file(f)
  rescue => err
    RubyReq::warn err.message
    RubyReq::warn err.backtrace.join("\n")
    exit 1 if RubyReq::method == 'strict'
  end
end
else
RubyReq::argv.each do |f|
  begin
    req += process_file(f)
  rescue SyntaxError => err
    RubyReq::warn err.message
    exit 1
  rescue => err
    RubyReq::warn err.message
    RubyReq::warn err.backtrace.join("\n")
    exit 1 if RubyReq::method == 'strict'
  end
end
end

req.sort.uniq.each do |r|
  rb, ext = RubyReq::base_ext(r)
  rb = r unless ext == '.rb'
  rn, rs = RubyReq::split_canon(rb)
  if rs[0] == 'rubygems'
    RubyReq::warn "forbidden dependency: #{rn}"
    exit 1 unless RubyReq::method == 'relaxed'
  end
  if $weak_prov.has_key?(rn)
    RubyReq::warn "#{rn} internally povided by #{$weak_prov[rn]}"
    next
  end
  if rs[0] == ''
    RubyReq::warn "absolute requires: #{rn}"
    exit 1 if RubyReq::method == 'strict'
    puts rn
  elsif rs.include?('..')
    RubyReq::warn "relative requires: #{rn}"
    exit 1 unless RubyReq::method == 'relaxed'
  elsif rs.include?('.')
    RubyReq::warn "dot-slash requires: #{rn}"
    exit 1 unless RubyReq::method == 'relaxed'
  else
    puts "ruby#{$major}(#{rn})"
  end
end

# vim: set ft=ruby ts=2 sts=2 sw=2 et:
